// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License in the LICENSE-APACHE file or at:
//     https://www.apache.org/licenses/LICENSE-2.0

//! Row / column solver

use std::marker::PhantomData;

use super::{
    AxisInfo, Direction, Margins, RowStorage, RowTemp, RulesSetter, RulesSolver, SizeRules,
};
use crate::geom::Rect;

/// A [`RulesSolver`] for rows (and, without loss of generality, for columns).
///
/// This is parameterised over:
///
/// -   `T:` [`RowTemp`] — temporary storage type
/// -   `S:` [`RowStorage`] — persistent storage type
pub struct RowSolver<T: RowTemp, S: RowStorage> {
    // Generalisation implies that axis.vert() is incorrect
    axis: AxisInfo,
    axis_is_vertical: bool,
    rules: SizeRules,
    widths: T,
    _s: PhantomData<S>,
}

impl<T: RowTemp, S: RowStorage> RowSolver<T, S> {
    /// Construct.
    ///
    /// - `axis`: `AxisInfo` instance passed into `size_rules`
    /// - `dim`: direction and number of items
    /// - `storage`: reference to persistent storage
    pub fn new<D: Direction>(axis: AxisInfo, dim: (D, usize), storage: &mut S) -> Self {
        let mut widths = T::default();
        widths.set_len(dim.1);
        assert!(widths.as_ref().iter().all(|w| *w == 0));
        storage.set_len(dim.1 + 1);

        let axis_is_vertical = axis.vertical ^ dim.0.is_vertical();

        if axis.has_fixed && axis_is_vertical {
            // TODO: cache this for use by set_rect?
            SizeRules::solve_seq(widths.as_mut(), storage.as_ref(), axis.other_axis);
        }

        RowSolver {
            axis,
            axis_is_vertical,
            rules: SizeRules::EMPTY,
            widths,
            _s: Default::default(),
        }
    }
}

impl<T: RowTemp, S: RowStorage> RulesSolver for RowSolver<T, S> {
    type Storage = S;
    type ChildInfo = usize;

    fn for_child<CR: FnOnce(AxisInfo) -> SizeRules>(
        &mut self,
        storage: &mut Self::Storage,
        child_info: Self::ChildInfo,
        child_rules: CR,
    ) {
        if self.axis.has_fixed && self.axis_is_vertical {
            self.axis.other_axis = self.widths.as_ref()[child_info];
        }
        let child_rules = child_rules(self.axis);
        if !self.axis_is_vertical {
            storage.as_mut()[child_info] = child_rules;
            self.rules += child_rules;
        } else {
            self.rules = self.rules.max(child_rules);
        }
    }

    fn finish<ColIter, RowIter>(
        self,
        storage: &mut Self::Storage,
        _: ColIter,
        _: RowIter,
    ) -> SizeRules
    where
        ColIter: Iterator<Item = (usize, usize, usize)>,
        RowIter: Iterator<Item = (usize, usize, usize)>,
    {
        let cols = storage.as_ref().len() - 1;
        if !self.axis_is_vertical {
            storage.as_mut()[cols] = self.rules;
        }

        self.rules
    }
}

/// A [`RulesSetter`] for rows (and, without loss of generality, for columns).
///
/// This is parameterised over:
///
/// -   `D:` [`Direction`] — whether this represents a row or a column
/// -   `T:` [`RowTemp`] — temporary storage type
/// -   `S:` [`RowStorage`] — persistent storage type
pub struct RowSetter<D, T: RowTemp, S: RowStorage> {
    crect: Rect,
    inter: u32,
    widths: T,
    direction: D,
    _s: PhantomData<S>,
}

impl<D: Direction, T: RowTemp, S: RowStorage> RowSetter<D, T, S> {
    pub fn new(mut rect: Rect, margins: Margins, dim: (D, usize), storage: &mut S) -> Self {
        let mut widths = T::default();
        widths.set_len(dim.1);
        storage.set_len(dim.1 + 1);

        rect.pos += margins.first;
        rect.size -= margins.first + margins.last;
        let mut crect = rect;

        let (width, inter) = if dim.0.is_horizontal() {
            crect.size.0 = 0; // hack to get correct first offset
            (rect.size.0, margins.inter.0)
        } else {
            crect.size.1 = 0;
            (rect.size.1, margins.inter.1)
        };

        SizeRules::solve_seq(widths.as_mut(), storage.as_ref(), width);

        RowSetter {
            crect,
            inter,
            widths,
            direction: dim.0,
            _s: Default::default(),
        }
    }
}

impl<D: Direction, T: RowTemp, S: RowStorage> RulesSetter for RowSetter<D, T, S> {
    type Storage = S;
    type ChildInfo = usize;

    fn child_rect(&mut self, child_info: Self::ChildInfo) -> Rect {
        if self.direction.is_horizontal() {
            self.crect.pos.0 += (self.crect.size.0 + self.inter) as i32;
            self.crect.size.0 = self.widths.as_ref()[child_info];
        } else {
            self.crect.pos.1 += (self.crect.size.1 + self.inter) as i32;
            self.crect.size.1 = self.widths.as_ref()[child_info];
        }
        self.crect
    }
}
